/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

#include "pgmrendersink.h"
#include "pgmrendermedia-iface.h"
#include <string.h>

static GstElementDetails pgm_render_sink_details =
  GST_ELEMENT_DETAILS ("Video Sink", "Sink/Video",
		       "Video output to PgmRender drawable",
		       "Loïc Molinari <loic@fluendo.com>");

static GstStaticPadTemplate pgm_render_sink_template =
  GST_STATIC_PAD_TEMPLATE ("sink", GST_PAD_SINK, GST_PAD_ALWAYS,
                           GST_STATIC_CAPS
                           ("video/x-raw-rgb, "
                            "framerate = (fraction) [ 0, MAX ], "
                            "width = (int) [ 1, MAX ], "
                            "height = (int) [ 1, MAX ]; "
                            "video/x-raw-yuv, "
                            "framerate = (fraction) [ 0, MAX ], "
                            "width = (int) [ 1, MAX ], "
                            "height = (int) [ 1, MAX ]"));

GST_DEBUG_CATEGORY_STATIC (gst_debug_pgmrendersink);
#define GST_CAT_DEFAULT gst_debug_pgmrendersink

enum {
  ARG_0,
  ARG_DRAWABLE
};

static void pgm_render_sink_set_property
  (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void pgm_render_sink_get_property
  (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void pgm_render_sink_get_times
  (GstBaseSink *bsink, GstBuffer *buf, GstClockTime *start, GstClockTime *end);
static GstFlowReturn pgm_render_sink_show (GstBaseSink *bsink, GstBuffer *buf);
static GstStateChangeReturn pgm_render_sink_change_state
  (GstElement *element, GstStateChange transition);
static void pgm_render_sink_dispose (GObject *object);
static void pgm_render_sink_finalize (GObject *object);
static GstCaps *pgm_render_sink_getcaps (GstBaseSink *bsink);
static gboolean pgm_render_sink_setcaps (GstBaseSink *bsink, GstCaps *caps);

static GstVideoSinkClass *parent_class = NULL;

static void
pgm_render_sink_set_property (GObject *object, guint prop_id,
			      const GValue *value, GParamSpec *pspec)
{
  PgmRenderSink *sink;

  g_return_if_fail (PGM_RENDER_IS_SINK (object));
  sink = PGM_RENDER_SINK (object);

  switch (prop_id)
    {
    case ARG_DRAWABLE:
      sink->drawable = PGM_RENDER_DRAWABLE (g_value_get_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
pgm_render_sink_get_property (GObject *object, guint prop_id, GValue *value,
			      GParamSpec *pspec)
{
  PgmRenderSink *sink;

  g_return_if_fail (PGM_RENDER_IS_SINK (object));
  sink = PGM_RENDER_SINK (object);

  switch (prop_id)
    {
    case ARG_DRAWABLE:
      g_value_set_object (value, sink->drawable);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
pgm_render_sink_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details (element_class, &pgm_render_sink_details);
  gst_element_class_add_pad_template (element_class,
				      gst_static_pad_template_get
				      (&pgm_render_sink_template));
}

static void
pgm_render_sink_class_init (PgmRenderSinkClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseSinkClass *gstbasesink_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstbasesink_class = (GstBaseSinkClass *) klass;

  parent_class = g_type_class_peek_parent (klass);
/*   parent_class = g_type_class_ref (GST_TYPE_VIDEO_SINK); */

  gobject_class->dispose = pgm_render_sink_dispose;
  gobject_class->finalize = pgm_render_sink_finalize;

  gstelement_class->change_state = pgm_render_sink_change_state;
  gobject_class->set_property = pgm_render_sink_set_property;
  gobject_class->get_property = pgm_render_sink_get_property;

  g_object_class_install_property
    (gobject_class, ARG_DRAWABLE,
     g_param_spec_object ("drawable", "Drawable",
			  "A PgmRender drawable on which the sink ouput " \
			  "video frames", PGM_RENDER_TYPE_DRAWABLE,
			  G_PARAM_READWRITE));

  gstbasesink_class->get_caps = GST_DEBUG_FUNCPTR (pgm_render_sink_getcaps);
  gstbasesink_class->set_caps = GST_DEBUG_FUNCPTR (pgm_render_sink_setcaps);
  gstbasesink_class->get_times = GST_DEBUG_FUNCPTR (pgm_render_sink_get_times);
  gstbasesink_class->preroll = GST_DEBUG_FUNCPTR (pgm_render_sink_show);
  gstbasesink_class->render = GST_DEBUG_FUNCPTR (pgm_render_sink_show);
}

static void
pgm_render_sink_init (PgmRenderSink *sink)
{
  sink->time = 0;
  sink->framerate_n = 0;
  sink->framerate_d = 0;
  sink->width = 0;
  sink->height = 0;
  sink->mutex = g_mutex_new ();
  sink->drawable = NULL;
}

static void
pgm_render_sink_dispose (GObject *object)
{
  return;
}

static void
pgm_render_sink_finalize (GObject *object)
{
  PgmRenderSink *sink = PGM_RENDER_SINK (object);

  g_mutex_free (sink->mutex);
}

static void
pgm_render_sink_get_times (GstBaseSink *bsink, GstBuffer *buf,
			   GstClockTime *start, GstClockTime *end)
{
  PgmRenderSink *sink = PGM_RENDER_SINK (bsink);

  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf))
    {
      *start = GST_BUFFER_TIMESTAMP (buf);

      if (GST_BUFFER_DURATION_IS_VALID (buf))
	*end = *start + GST_BUFFER_DURATION (buf);
      else
	if (sink->framerate_n > 0)
	  *end = *start + gst_util_uint64_scale_int (GST_SECOND, 
						     sink->framerate_d,
						     sink->framerate_n);
    }
}

static GstFlowReturn
pgm_render_sink_show (GstBaseSink *bsink, GstBuffer *buf)
{
  PgmRenderSink *sink = PGM_RENDER_SINK (bsink);
  GstClockTime buf_time;

  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf))
    sink->time = GST_BUFFER_TIMESTAMP (buf);

  buf_time = sink->time;

  if (GST_BUFFER_DURATION_IS_VALID (buf))
    sink->time += GST_BUFFER_DURATION (buf);
  else if (sink->framerate_n > 0)
    sink->time += gst_util_uint64_scale_int (GST_SECOND, sink->framerate_d,
					     sink->framerate_n);

  if (sink->drawable && buf)
    pgm_render_media_set_gst_image (PGM_RENDER_MEDIA_IFACE (sink->drawable),
				    sink->format, sink->width,
				    sink->height, sink->stride, buf);

  return GST_FLOW_OK;
}

static GstCaps *
pgm_render_sink_getcaps (GstBaseSink *bsink)
{
  PgmRenderSink *sink = PGM_RENDER_SINK (bsink);
  PgmRenderDrawableClass *klass;

  if (!sink->drawable)
    goto copy;

  klass = PGM_RENDER_DRAWABLE_GET_CLASS (sink->drawable);

  if (klass->caps)
    return gst_caps_ref (klass->caps);

 copy:
  return gst_caps_copy (gst_pad_get_pad_template_caps
                        (GST_VIDEO_SINK_PAD (sink)));
}

static gboolean 
pgm_render_sink_setcaps (GstBaseSink *bsink, GstCaps *caps)
{
  PgmRenderSink *sink = PGM_RENDER_SINK (bsink);
  gboolean supported = FALSE;
  gint i = 0;
  gint fourcc, red_mask, green_mask, blue_mask, alpha_mask, depth, bpp;
  GstStructure *structure;
  gboolean ret;
  guint32 format;

  /* The fourcc/format table */
  static FourccList fourcc_list[] =
    {
      { "I420", PGM_RENDER_MEDIA_I420 },
      { "IYUV", PGM_RENDER_MEDIA_I420 },
      { "YV12", PGM_RENDER_MEDIA_YV12 },
      { "YUYV", PGM_RENDER_MEDIA_YUYV },
      { "YUY2", PGM_RENDER_MEDIA_YUYV },
      { "V422", PGM_RENDER_MEDIA_YUYV },
      { "YUNV", PGM_RENDER_MEDIA_YUYV },
      { "UYVY", PGM_RENDER_MEDIA_UYVY },
      { "Y422", PGM_RENDER_MEDIA_UYVY },
      { "UYNV", PGM_RENDER_MEDIA_UYVY },
      { NULL, 0 }
    };

  /* The rgb/format table */
  static RgbList rgb_list[] =
    {
      { "RGB", 24, 24, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000,
        PGM_RENDER_MEDIA_RGB },
      { "BGR", 24, 24, 0x000000ff, 0x0000ff00, 0x00ff0000, 0x00000000,
        PGM_RENDER_MEDIA_BGR },
      { "RGBA", 32, 32, 0xff000000, 0x00ff0000, 0x0000ff00, 0x000000ff,
        PGM_RENDER_MEDIA_RGBA },
      { "BGRA", 32, 32, 0x0000ff00, 0x00ff0000, 0xff000000, 0x000000ff,
        PGM_RENDER_MEDIA_BGRA, },
      { NULL, 0, 0, 0x0, 0x0, 0x0, 0x0, 0 }
    };

  structure = gst_caps_get_structure (caps, 0);

  ret = gst_structure_get_int (structure, "width", &sink->width);
  ret &= gst_structure_get_int (structure, "height", &sink->height);
  ret &= gst_structure_get_fraction (structure, "framerate",
				     &sink->framerate_n, &sink->framerate_d);

  if (!ret)
    {
      g_print ("pgm_render_sink_setcaps(): Incomplete caps structure\n");
      return FALSE;
    }

  /* YUV formats */
  if (!strcmp (gst_structure_get_name (structure), "video/x-raw-yuv"))
    {
      ret = gst_structure_get_fourcc (structure, "format", &format);
      if (!ret)
      {
        g_print ("pgm_render_sink_setcaps(): Incomplete caps structure\n");
        return FALSE;
      }

      /* Check if we support the format */
      while (fourcc_list[i].fourcc && !supported) 
        {
          fourcc = GST_MAKE_FOURCC
            (fourcc_list[i].fourcc[0], fourcc_list[i].fourcc[1],
             fourcc_list[i].fourcc[2], fourcc_list[i].fourcc[3]);
          if (fourcc == format)
            supported = TRUE;
          else
            i++;
        }

      if (!supported)
        {
          g_print ("pgm_render_sink_setcaps(): Unsupported YUV format.\n");
          return FALSE;
        }

      sink->format = fourcc_list[i].format;
/*       g_print ("YUV stream. size=%dx%d fps=%d/%d format=%d\n", */
/*                sink->width, sink->height, sink->framerate_n, */
/*                sink->framerate_d, sink->format); */
    }

  /* RGB formats */
  else if (!strcmp (gst_structure_get_name (structure), "video/x-raw-rgb"))
    {
      ret = gst_structure_get_int (structure, "red_mask", &red_mask);
      ret &= gst_structure_get_int (structure, "green_mask", &green_mask);
      ret &= gst_structure_get_int (structure, "blue_mask", &blue_mask);
      ret &= gst_structure_get_int (structure, "depth", &depth);
      ret &= gst_structure_get_int (structure, "bpp", &bpp);
      if (depth == 32)
        {
          ret &= gst_structure_get_int (structure, "alpha_mask", &alpha_mask);
          ret &= (alpha_mask != 0);
        }
      else
        alpha_mask = 0;
      if (!ret)
        {
          g_print ("pgm_render_sink_setcaps(): Incomplete caps structure\n");
          return FALSE;
        }

      /* Check if we support the format */
      while (rgb_list[i].rgb && !supported)
        {
          if (rgb_list[i].red_mask == red_mask
              && rgb_list[i].green_mask == green_mask
              && rgb_list[i].blue_mask == blue_mask
              && (alpha_mask == 0 || rgb_list[i].alpha_mask == alpha_mask)
              && rgb_list[i].depth == depth
              && rgb_list[i].bpp == bpp)
            supported = TRUE;
          else
            i++;
        }

      if (!supported)
        {
          g_print ("pgm_render_sink_setcaps(): Unsupported RGB format.\n");
          return FALSE;
        }

      sink->format = rgb_list[i].format;
/*       g_print ("RGB stream. size=%dx%d fps=%d/%d format=%d\n", */
/*                sink->width, sink->height, sink->framerate_n, */
/*                sink->framerate_d, sink->format); */
    }

  return ret;
}

static GstStateChangeReturn
pgm_render_sink_change_state (GstElement *element, GstStateChange transition)
{
  PgmRenderSink *sink;

  sink = PGM_RENDER_SINK (element);

  switch (transition)
    {
    case GST_STATE_CHANGE_READY_TO_PAUSED:
      sink->time = 0; 
      break;
    default:
      break;
    }

  return GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
}

GType
pgm_render_sink_get_type (void)
{
  static GType sink_type = 0;

  if (G_UNLIKELY (!sink_type))
    {
      static const GTypeInfo sink_info =
	{
	  sizeof (PgmRenderSinkClass),
	  pgm_render_sink_base_init,
	  NULL,
	  (GClassInitFunc) pgm_render_sink_class_init,
	  NULL,
	  NULL,
	  sizeof (PgmRenderSink),
	  0,
	  (GInstanceInitFunc) pgm_render_sink_init,
	};

      sink_type = g_type_register_static (GST_TYPE_VIDEO_SINK, "PgmRenderSink",
					  &sink_info, 0);
    }

  return sink_type;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  GST_DEBUG_CATEGORY_INIT (gst_debug_pgmrendersink, "pgmrendersink", 0,
      "pgmrendersink element");

  return gst_element_register (plugin, "pgmrendersink", GST_RANK_NONE,
			       PGM_RENDER_TYPE_SINK);
}

/* FIXME: Include package/origin/version with the autotools */
#define PACKAGE ""
GST_PLUGIN_DEFINE (0, 10, "pgmrendersink",
		   "Outputs a video to a PgmRender drawable object",
		   plugin_init, "0.1", "LGPL", "Pigment", "")
