/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrenderbasewindow
 * @short_description: Base class for Window.
 *
 * #PgmRenderBaseWindow is a base class for Window render class.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrenderbasewindow.h"

static void pgm_render_base_window_class_init (PgmRenderBaseWindowClass *klass);
static void pgm_render_base_window_init (PgmRenderBaseWindow *object);
static void pgm_render_base_window_finalize (GObject *object);

static GObjectClass *parent_class = NULL;

G_DEFINE_ABSTRACT_TYPE (PgmRenderBaseWindow, pgm_render_base_window,
			G_TYPE_OBJECT);

static void
pgm_render_base_window_class_init (PgmRenderBaseWindowClass *klass)
{
  GObjectClass *gobject_class = (GObjectClass *) klass;

  parent_class = g_type_class_peek_parent (klass);
  gobject_class->finalize = pgm_render_base_window_finalize;
}

static void
pgm_render_base_window_init (PgmRenderBaseWindow *object)
{
  object->lock = g_new0 (GStaticMutex, 1);
  g_static_mutex_init (object->lock);
  object->dirty = TRUE;
  object->far_drawable = NULL;
  object->z_ordered_drawable = NULL;
  object->near_drawable = NULL;
  object->width = 0;
  object->height = 0;
  object->canvas_width = 0;
  object->canvas_height = 0;
  object->sx = 1.0f;
  object->sy = 1.0f;
  object->scale_mode = PGM_RENDER_WINDOW_ADJUST;
/*   object->width_aspect_factor = 1.0f; */
/*   object->height_aspect_factor = 1.0f; */
}

static void
pgm_render_base_window_finalize (GObject *object)
{
  PgmRenderBaseWindow *window = PGM_RENDER_BASE_WINDOW (object);

  g_static_mutex_free (window->lock);
  g_free (window->lock);
  window->lock = NULL;

  if (window->far_drawable)
    {
      /* FIXME: Free all the drawables */
      g_list_free (window->far_drawable);
      window->far_drawable = NULL;
    }
  if (window->z_ordered_drawable)
    {
      /* FIXME: Free all the drawables */
      g_list_free (window->z_ordered_drawable);
      window->z_ordered_drawable = NULL;
    }
  if (window->near_drawable)
    {
      /* FIXME: Free all the drawables */
      g_list_free (window->near_drawable);
      window->near_drawable = NULL;
    }

  G_OBJECT_CLASS(parent_class)->finalize (object);
}

/**
 * pgm_render_base_window_new:
 *
 * Creates a new #PgmRenderBaseWindow instance.
 *
 * Returns: A new #PgmRenderBaseWindow.
 */
PgmRenderBaseWindow *
pgm_render_base_window_new (void)
{
  return PGM_RENDER_BASE_WINDOW
    (g_object_new (PGM_RENDER_TYPE_BASE_WINDOW, NULL));
}

/**
 * pgm_render_base_window_update:
 * @window: A #PgmRenderBaseWindow object.
 *
 * Update the window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_update (PgmRenderBaseWindow *window)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->update)
    klass->update (window);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_cursor:
 * @window: A #PgmRenderBaseWindow object.
 * @cursor: The cursor to set.
 *
 * Set the #cursor of the window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_cursor (PgmRenderBaseWindow *window, PgmRenderCursor cursor)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window), PGM_RENDER_ERROR_X);

  window->cursor = cursor;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_cursor)
    klass->set_cursor (window, cursor);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_cursor:
 * @window: A #PgmRenderBaseWindow object.
 * @cursor: The cursor to fill.
 *
 * Retrieves the #cursor of the window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_cursor (PgmRenderBaseWindow *window, PgmRenderCursor *cursor)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window), PGM_RENDER_ERROR_X);
  g_return_val_if_fail (cursor, PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  *cursor = window->cursor;

  if (klass->get_cursor)
    klass->get_cursor (window, cursor);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_color:
 * @window: A #PgmRenderBaseWindow object.
 * @r: The R color layer.
 * @g: The G color layer.
 * @b: The B color layer.
 * @a: The A color layer.
 *
 * Sets the background color of #window to (#r, #g, #b, #a).
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_base_window_set_color (PgmRenderBaseWindow *window, guchar r,
				  guchar g, guchar b, guchar a)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_color)
    klass->set_color (window, r, g, b, a);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_color:
 * @window: A #PgmRenderBaseWindow object.
 * @r: The R color layer to fill.
 * @g: The G color layer to fill.
 * @b: The B color layer to fill.
 * @a: The A color layer to fill.
 *
 * Retrieves the background color of #window.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_base_window_get_color (PgmRenderBaseWindow *window, guchar *r,
				  guchar *g, guchar *b, guchar *a)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_color)
    klass->get_color (window, r, g, b, a);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_title:
 * @window: A #PgmRenderBaseWindow object.
 * @window_title: The window title.
 * @icon_title: The icon title.
 *
 * Set the titles appearing in the title bar and in the iconified
 * window list.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_title (PgmRenderBaseWindow *window,
				  const gchar *window_title,
				  const gchar *icon_title)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_title)
    klass->set_title (window, window_title, icon_title);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_size:
 * @window: A #PgmRenderBaseWindow object.
 * @width: The window width.
 * @height: The window height.
 *
 * Set #width and #height of the window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_size (PgmRenderBaseWindow *window, gint width,
				 gint height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  window->width = width;
  window->height = height;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);
  
  if (klass->set_size)
    klass->set_size (window, width, height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_size:
 * @window: A #PgmRenderBaseWindow object.
 * @width: The width to fill.
 * @height: The height to fill.
 *
 * Retrieves the size (#width, #height) of the window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_size (PgmRenderBaseWindow *window, gint *width,
				 gint *height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  *width = window->width;
  *height = window->height;

  if (klass->get_size)
    klass->get_size (window, width, height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_canvas_size:
 * @window: A #PgmRenderBaseWindow object.
 * @canvas_width: The canvas width.
 * @canvas_height: The canvas height.
 *
 * Set the (#canvas_width, #canvas_height) canvas size of the #window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_canvas_size (PgmRenderBaseWindow *window,
                                        gint canvas_width, gint canvas_height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  window->canvas_width = canvas_width;
  window->canvas_height = canvas_height;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);
  
  if (klass->set_canvas_size)
    klass->set_canvas_size (window, canvas_width, canvas_height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_canvas_size:
 * @window: A #PgmRenderBaseWindow object.
 * @canvas_width: The canvas width to fill.
 * @canvas_height: The canvas height to fill.
 *
 * Retrieves the (#canvas_width, #canvas_height) canvas size of the #window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_canvas_size (PgmRenderBaseWindow *window,
                                        gint *canvas_width,
                                        gint *canvas_height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (canvas_width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (canvas_height, PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  *canvas_width = window->canvas_width;
  *canvas_height = window->canvas_height;

  if (klass->get_canvas_size)
    klass->get_canvas_size (window, canvas_width, canvas_height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_fullscreen:
 * @window: A #PgmRenderBaseWindow object.
 * @fullscreen: The fullscreen state.
 * @width: The width size of the fullscreen mode.
 * @height: The height size of the fullscreen mode.
 *
 * Set/unset fullscreen mode in the specified #width/#height mode.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_fullscreen (PgmRenderBaseWindow *window,
				       gboolean fullscreen, gint width,
                                       gint height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_fullscreen)
    klass->set_fullscreen (window, fullscreen, width, height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_fullscreen:
 * @window: A #PgmRenderBaseWindow object.
 * @fullscreen: The fullscreen state to fill.
 *
 * Retrieves the fullscreen mode.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_fullscreen (PgmRenderBaseWindow *window,
				       gboolean *fullscreen)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  *fullscreen = FALSE;

  if (klass->get_fullscreen)
    klass->get_fullscreen (window, fullscreen);
  
  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_scale:
 * @window: A #PgmRenderBaseWindow object.
 * @sx: Scale factor for the x dimension. 1.0 by default.
 * @sy: Scale factor for the y dimension. 1.0 by default.
 *
 * Set the (#sx, #sy) scale factor of the canvas dimensions.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_scale (PgmRenderBaseWindow *window, gfloat sx,
                                  gfloat sy)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  window->sx = sx;
  window->sy = sy;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_scale)
    klass->set_scale (window, sx, sy);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_scale:
 * @window: A #PgmRenderBaseWindow object.
 * @sx: The scale factor for the x dimension to fill.
 * @sy: The scale factor for the y dimension to fill.
 *
 * Retrieves the (#sx, #sy) scale factor of the canvas dimensions.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_scale (PgmRenderBaseWindow *window, gfloat *sx,
                                  gfloat *sy)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (sx, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (sy, PGM_RENDER_ERROR_X);

  *sx = window->sx;
  *sy = window->sy;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_scale)
    klass->get_scale (window, sx, sy);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_screen_resolution:
 * @window: A #PgmRenderBaseWindow object.
 * @width: The width in pixels to fill.
 * @height: The height in pixels to fill.
 *
 * Retrieves the (#width, #height) screen resolution in pixels of the
 * #window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_screen_resolution (PgmRenderBaseWindow *window,
                                              gint *width, gint *height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_screen_resolution)
    klass->get_screen_resolution (window, width, height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_screen_size:
 * @window: A #PgmRenderBaseWindow object.
 * @width: The width in pixels to fill.
 * @height: The height in pixels to fill.
 *
 * Retrieves the (#width, #height) screen size in pixels of the
 * #window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_screen_size (PgmRenderBaseWindow *window,
					gint *width, gint *height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_screen_size)
    klass->get_screen_size (window, width, height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_screen_size_mm:
 * @window: A #PgmRenderBaseWindow object.
 * @width: The width in millimeters to fill.
 * @height: The height in millimeters to fill.
 *
 * Retrieves the physical (#width, #height) screen size in millimeters
 * of the #window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_screen_size_mm (PgmRenderBaseWindow *window,
                                           gint *width, gint *height)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_screen_size_mm)
    klass->get_screen_size_mm (window, width, height);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_vertical_sync:
 * @window: A #PgmRenderBaseWindow object.
 * @vsync: The vertical synchronization state. TRUE by default.
 *
 * Set the vertical synchronization state of the @window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_vertical_sync (PgmRenderBaseWindow *window,
					  gboolean vsync)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_vertical_sync)
    klass->set_vertical_sync (window, vsync);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_vertical_sync:
 * @window: A #PgmRenderBaseWindow object.
 * @vsync: The vertical synchronization state to fill.
 *
 * Retrieves the vertical synchronization state of the @window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_vertical_sync (PgmRenderBaseWindow *window,
					  gboolean *vsync)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_vertical_sync)
    klass->get_vertical_sync (window, vsync);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_events:
 * @window: A #PgmRenderBaseWindow object.
 * @events: The event list to fill. Must be freed after usage.
 *
 * Retrieves the generated events since the last get_events.
 *
 * Returns: An error status.
 *
 * MT safe.
 */
PgmRenderError
pgm_render_base_window_get_events (PgmRenderBaseWindow *window, GList **events)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_events)
    klass->get_events (window, events);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_start_loop:
 * @window: A #PgmRenderBaseWindow object.
 *
 * Runs the main loop until #pgm_render_base_window_quit_loop() is
 * called.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_start_loop (PgmRenderBaseWindow *window)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->start_loop)
    klass->start_loop (window);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_quit_loop:
 * @window: A #PgmRenderBaseWindow object.
 *
 * Stops the main loop.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_quit_loop (PgmRenderBaseWindow *window)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->quit_loop)
    klass->quit_loop (window);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_add_drawable:
 * @window: A #PgmRenderBaseWindow object.
 * @drawable: The #PgmRenderDrawable object to add.
 * @order: The ordering type.
 *
 * Add a #PgmRenderDrawable object to the #window with the given
 * #order type.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_add_drawable (PgmRenderBaseWindow *window,
				     PgmRenderDrawable *drawable,
				     PgmRenderDrawableOrder order)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->add_drawable)
    klass->add_drawable (window, drawable, order);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_del_drawable:
 * @window: A #PgmRenderBaseWindow object.
 * @drawable: The #PgmRenderDrawable object to del.
 *
 * Del a #PgmRenderDrawable object from the #window.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_del_drawable (PgmRenderBaseWindow *window,
				     PgmRenderDrawable *drawable)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->del_drawable)
    klass->del_drawable (window, drawable);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_reorder_drawable:
 * @window: A #PgmRenderBaseWindow object.
 * @drawable: The #PgmRenderDrawable object to reorder.
 * @position: the new position of the object.
 *
 * Reorder a #PgmRenderDrawable object in the #window at the given
 * #position. A #position of -1 put the drawable at the end.
 *
 * Returns: An error status.
 *
 * FIXME: Not needed anymore
 */
PgmRenderError
pgm_render_base_window_reorder_drawable (PgmRenderBaseWindow *window,
					 PgmRenderDrawable *drawable,
					 gint position)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (PGM_RENDER_IS_DRAWABLE (drawable), PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->del_drawable)
    klass->reorder_drawable (window, drawable, position);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_regenerate_drawables:
 * @window: A #PgmRenderBaseWindow object.
 *
 * Affect all the drawables of the #window which need to be displayed
 * at a fixed size. It regenerates those rasterized pixmaps which have
 * been affected by the #window scale factors.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_regenerate_drawables (PgmRenderBaseWindow *window)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->regenerate_drawables)
    klass->regenerate_drawables (window);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_set_scale_mode:
 * @window: A #PgmRenderBaseWindow object.
 * @scale_mode: The scaling mode to apply.
 *
 * Set the scaling mode to apply to the canvas size.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_set_scale_mode (PgmRenderBaseWindow *window,
                                       PgmRenderWindowScaleMode scale_mode)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);

  window->scale_mode = scale_mode;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->set_scale_mode)
    klass->set_scale_mode (window, scale_mode);

  return PGM_RENDER_ERROR_X;
}

/**
 * pgm_render_base_window_get_scale_mode:
 * @window: A #PgmRenderBaseWindow object.
 * @scale_mode: The scaling mode to fill.
 *
 * Retrieve the actual scaling mode applied to the canvas size.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_base_window_get_scale_mode (PgmRenderBaseWindow *window,
                                       PgmRenderWindowScaleMode *scale_mode)
{
  PgmRenderBaseWindowClass *klass;

  g_return_val_if_fail (PGM_RENDER_IS_BASE_WINDOW (window),
			PGM_RENDER_ERROR_X);
  g_return_val_if_fail (scale_mode, PGM_RENDER_ERROR_X);

  *scale_mode = window->scale_mode;

  klass = PGM_RENDER_BASE_WINDOW_GET_CLASS (window);

  if (klass->get_scale_mode)
    klass->get_scale_mode (window, scale_mode);

  return PGM_RENDER_ERROR_X;
}
